import time
import json
from pathlib import Path
from typing import TypeGuard, Any
from collections.abc import Iterator, Callable

import click

from .log import logger
from .sources.model import FeedItem
from .blur import Blurred


@click.group()
def main():
    """My Personal Media Feed"""


BLUE = (83, 158, 206)


def _check_source(source: Any) -> TypeGuard[Callable[[], Iterator[FeedItem]]]:
    if not callable(source):
        click.echo(f"{source} is not callable, ignoring source", err=True)
        return False
    return True


def _sources() -> Iterator[Callable[[], Iterator[FeedItem]]]:
    try:
        from my.config.feed import sources  # type: ignore[import]
    except Exception:
        click.echo(
            "Could not import sources from my.config.feed, see docs or https://github.com/purarue/dotfiles/blob/master/.config/my/my/config/feed.py as an example",
            err=True,
        )
        return

    assert callable(sources), "sources imported from my.config.feed is not a function"
    for src in iter(sources()):
        if _check_source(src):
            yield src


def data(
    *, allow: list[str], deny: list[str], blurred: Blurred | None, echo: bool = False
) -> Iterator[FeedItem]:
    for producer in _sources():
        func = f"{producer.__module__}.{producer.__qualname__}"
        if len(allow) > 0 and not any(substr in func for substr in allow):
            continue
        if len(deny) > 0 and any(substr in func for substr in deny):
            continue
        emitted: set[str] = set()
        start_time = time.time()
        func = f"{producer.__module__}.{producer.__qualname__}"
        ext = f"Extracting {click.style(func, fg='green')}"
        click.echo(f"{ext}...")
        for item in producer():
            assert isinstance(item, FeedItem)
            item.check()
            if item.id in emitted:
                logger.warning(f"Duplicate id: {item.id} {item}")
                continue
            emitted.add(item.id)
            if echo:
                print(item)
            if blurred and blurred.should_be_blurred(feed_item=item):
                item.blur()
                click.echo(f"Blurred image: {item.id=} {item.title=} {item.image_url=}")
            yield item
        took = time.time() - start_time
        click.echo(
            f"{ext}: {click.style(str(len(emitted)), fg=BLUE)} items (took {click.style(round(took, 2), fg=BLUE)} seconds)",
            err=True,
        )


# TODO: this could allow either passing the ID or the URL
def _parse_blur_file(
    ctx: click.Context, param: click.Parameter, value: Path | None
) -> Blurred | None:
    if value is not None:
        return Blurred.parse_file(value)
    return None


def _parse_sources(
    ctx: click.Context, param: click.Parameter, value: str | None
) -> list[str]:
    if value is not None:
        return [p.strip() for p in value.strip().split(",")]
    return []


@main.command(name="index", short_help="recompute feed data")
@click.option(
    "--echo/--no-echo",
    default=False,
    is_flag=True,
    help="Print feed items as they're computed",
)
@click.option(
    "-i",
    "--include-sources",
    default=None,
    help="A comma delimited list of substrings of sources to include. e.g. 'mpv,trakt,listens'",
    callback=_parse_sources,
    envvar="MY_FEED_INCLUDE_SOURCES",
)
@click.option(
    "-e",
    "--exclude-sources",
    default=None,
    envvar="MY_FEED_EXCLUDE_SOURCES",
    help="A comma delimited list of substrings of sources to exclude. e.g. 'mpv,trakt,listens'",
    callback=_parse_sources,
)
@click.option(
    "-E",
    "--exclude-id-file",
    default=None,
    help="A json file containing a list of IDs to exclude, from the /data/ids endpoint. reduces amount of data to sync to the server",
    type=click.Path(exists=True, path_type=Path),
)
@click.option(
    "-C",
    "--write-count-to",
    type=click.Path(writable=True, path_type=Path),
    default=None,
    help="Write the number of items to this file",
)
@click.option(
    "-B",
    "--blur-images-file",
    "blurred",
    default=None,
    help="A file containing a list of image URLs to blur, one per line",
    type=click.Path(exists=True, path_type=Path),
    callback=_parse_blur_file,
)
@click.argument(
    "OUTPUT", type=click.Path(writable=True, path_type=Path), required=False
)
def index(
    echo: bool,
    include_sources: list[str],
    exclude_sources: list[str],
    write_count_to: Path | None,
    output: Path | None,
    blurred: Blurred | None,
    exclude_id_file: Path | None,
) -> None:
    if blurred:
        click.echo("Blurred matchers:")
        click.echo("\n".join(map(str, blurred.items)))

    exclude_ids: set[str] = set()
    if exclude_id_file is not None:
        click.echo(f"Reading exclude IDs from '{exclude_id_file}'")
        exclude_ids = set(json.loads(exclude_id_file.read_text()))
    all_items = list(
        data(allow=include_sources, deny=exclude_sources, blurred=blurred, echo=echo)
    )
    items = [i for i in all_items if i.id not in exclude_ids]

    if exclude_ids:
        click.echo(f"Excluded {click.style(len(all_items) - len(items), BLUE)} items")
    click.echo(f"Total: {click.style(len(items), BLUE)} items")
    if output is not None:
        if len(items) == 0:
            click.echo("No items to sync!", err=True)
        else:
            click.echo(f"Writing to '{output}'", err=True)
            with output.open("w") as f:
                for item in items:
                    f.write(item.to_json())
                    f.write("\n")
        if write_count_to:
            write_count_to.write_text(str(len(items)))

    # hm: consume the rest of the generator so that cachew db closes...?
    #
    # happens when I use the albums module since it uses via_location.localize
    #
    # so, I suppose when trying to exit, since we never finished using the generator,
    # its still trying to write, but we exit so it fails to write to the closed db
    #
    # for posterity: https://gist.github.com/purarue/58f0b9f9dbcfeb970270e7fba99604d7
    # try:
    #     from my.time.tz.via_location import loc_tz_getter
    #
    #     list(loc_tz_getter())  # type: ignore
    # except Exception as e:
    #     logger.exception(e)


if __name__ == "__main__":
    main(prog_name="my_feed")



"""
Allows custom transformations to be applied to feed items

After creating a list called TRANSFORMS in my.config.feed, you can, you can
wrap some source with transform(source), e.g.:

def _fix_artist_name(item: FeedItem) -> Optional[FeedItem]:
    if item.ftype != 'mpv':
        return item

    if item.creator = "something I want to ignore":
        return None  # drop item

    # e.g. fix a misspelled artist name or one that's slightly
    # different from the one in the musicbrainz database
    if item.creator = "...":
        data = dataclasses.asdict(item)
        data['creator'] = "something"
        return FeedItem(**data)
    return item  # otherwise return as is


TRANSFORMS = [_fix_artist_name]

def sources() -> Iterator[Callable[[], Iterator["FeedItem"]]]:
    from my_feed.transform import transform

    yield transform(mpv.history)
"""

import functools
from typing import Optional
from collections.abc import Callable, Iterator

from .sources.model import FeedItem
from .log import logger


# if this returns nothing, the item is dropped
# if you don't want to transform the item, yield the item itself
TransformFunction = Callable[[FeedItem], Optional[FeedItem]]

TRANSFORMS: list[TransformFunction] = []

try:
    from my.config.feed import TRANSFORMS as MY_TRANSFORMS  # type: ignore[import]

    assert isinstance(MY_TRANSFORMS, list)
    for tr in MY_TRANSFORMS:
        assert callable(tr)
    TRANSFORMS.extend(MY_TRANSFORMS)

except ImportError as e:
    logger.debug(e, exc_info=True)


def transform(
    feed: Callable[[], Iterator[FeedItem]],
    transforms: list[TransformFunction] = TRANSFORMS,
) -> Callable[[], Iterator[FeedItem]]:
    """
    Receives a callable source as input, and wraps it, returning a callable
    This is the entrypoint to this module
    """

    @functools.wraps(feed)
    def _tr() -> Iterator[FeedItem]:
        yield from _transform(feed(), transforms)

    return _tr


def _transform(
    feed: Iterator[FeedItem], transforms: list[TransformFunction] = TRANSFORMS
) -> Iterator[FeedItem]:
    for item in feed:
        # update scope with item
        updated: FeedItem = item
        for transform in transforms:
            if transformed := transform(item):
                updated = transformed
            else:
                # drop item, was none
                break
        else:
            # we hit this block if we exhausted the transforms without breaking
            # so we yield the item here
            yield updated



import dataclasses
from datetime import date, timedelta, datetime


from .sources.model import FeedItem
from .log import logger


class Timeshift:
    def __init__(
        self, *, for_feeditems: set[str], timeshift_between: tuple[date, date]
    ):
        self.for_feeditems = for_feeditems
        self.earliest_start_year = date(1940, 1, 1)
        (
            self.started_watching_media,
            self.account_created,
        ) = timeshift_between  # 2000, 2016

    def matches(self, feeditem: FeedItem) -> bool:
        return (
            feeditem.ftype in self.for_feeditems
            and feeditem.when is not None
            and feeditem.when.date() < self.account_created
        )

    def _determine_timeshift(self, feeditem: FeedItem) -> date | None:
        # shift items before the account creation date to somewhere between the media start and end date
        # leave items after the account creation date alone

        # something released in 1920 -> 2005
        # something released in 1925 -> 2006
        # something released in 1980 -> 2009
        # something released in 2000 -> 2014
        # something released in 2016 -> 2016
        # something released in 2017 -> None

        if feeditem.when.date() < self.account_created:
            if feeditem.when.date() < self.earliest_start_year:
                return self.started_watching_media

            # feeditem date is larger than min date, this is the number of days between the two
            numerator: timedelta = feeditem.when.date() - self.earliest_start_year
            denominator: timedelta = self.account_created - self.earliest_start_year

            # frac is the fraction of the way between the min and max date
            frac = numerator.days / denominator.days
            assert 1 >= frac >= 0

            # scale the fraction to the number of days between started_watching_media and account_created
            # and add that to started_watching_media
            add_to_diff: timedelta = frac * (
                self.account_created - self.started_watching_media
            )
            shifted_date: date = self.started_watching_media + add_to_diff
            return shifted_date

        else:
            return None

    def timeshift(self, feeditem: FeedItem) -> FeedItem | None:
        if new_date := self._determine_timeshift(feeditem):
            logger.debug(
                f"timeshift {feeditem.ftype} {feeditem.title} from {feeditem.when.date()} to {new_date}"
            )
            data = dataclasses.asdict(feeditem)
            data["when"] = datetime.combine(
                new_date, feeditem.when.time(), tzinfo=feeditem.when.tzinfo
            )
            return FeedItem(**data)
        return None



from __future__ import annotations
import re
import fnmatch

from pathlib import Path
from enum import Enum
from typing import NamedTuple, TextIO

from .sources.model import FeedItem


class Attr(str, Enum):
    ID_FNMATCH = "id"
    TITLE_FNMATCH = "title"
    IMAGE_FNMATCH = "image_url"
    ID_REGEX = "id_regex"
    TITLE_REGEX = "title_regex"
    IMAGE_REGEX = "image_url_regex"

    @classmethod
    def from_str(cls, data: str) -> Attr:
        match data:
            case "id":
                return cls.ID_FNMATCH
            case "title":
                return cls.TITLE_FNMATCH
            case "image_url":
                return cls.IMAGE_FNMATCH
            case "id_regex":
                return cls.ID_REGEX
            case "title_regex":
                return cls.TITLE_REGEX
            case "image_url_regex":
                return cls.IMAGE_REGEX
            case _:
                raise ValueError(f"Unknown attribute: {data}")


class Blur(NamedTuple):
    attr: Attr
    pattern: str

    def __str__(self) -> str:
        return f"{self.attr}: {self.pattern}"

    @classmethod
    def from_str(cls, line: str) -> Blur:
        attr, _, value = line.partition(":")
        val = value.strip()
        if not val:
            raise ValueError(f"Empty value for attribute {attr}, line {line}")
        return cls(Attr.from_str(attr.strip().lower()), val)


class Blurred(NamedTuple):
    items: set[Blur]

    @classmethod
    def parse_blob(cls, f: TextIO, /) -> Blurred:
        return cls({Blur.from_str(line) for line in f if line.strip()})

    @classmethod
    def parse_file(cls, path: Path, /) -> Blurred:
        with path.open("r") as f:
            return cls.parse_blob(f)

    def should_be_blurred(self, *, feed_item: FeedItem) -> bool:
        for blur in self.items:
            if blur.attr == Attr.ID_FNMATCH:
                if fnmatch.fnmatch(feed_item.id, blur.pattern):
                    return True
            elif blur.attr == Attr.ID_REGEX:
                if re.search(blur.pattern, feed_item.id):
                    return True
            elif blur.attr == Attr.TITLE_FNMATCH:
                if fnmatch.fnmatch(feed_item.title, blur.pattern):
                    return True
            elif blur.attr == Attr.TITLE_REGEX:
                if re.search(blur.pattern, feed_item.title):
                    return True
            elif blur.attr in (Attr.IMAGE_FNMATCH, Attr.IMAGE_REGEX):
                if not feed_item.image_url:
                    continue
                if blur.attr == Attr.IMAGE_REGEX:
                    if re.search(blur.pattern, feed_item.image_url):
                        return True
                else:
                    if fnmatch.fnmatch(feed_item.image_url, blur.pattern):
                        return True
        return False



from my_feed.sources.trakt.tmdb import BASE_URL, _matches_trakt


def test_matches() -> None:
    assert _matches_trakt(BASE_URL + "/movie/145")
    assert _matches_trakt(BASE_URL + "/tv/523/season/2/episode/5")
    assert _matches_trakt(BASE_URL + "/tv/19481239/season/4912")
    assert _matches_trakt(BASE_URL + "/tv/423849")
    assert not _matches_trakt("https://something.org/tv/423849")
    assert not _matches_trakt(BASE_URL)


from my_feed.blur import Blurred, Attr, Blur


def test_parse_blob() -> None:
    from io import StringIO

    buf = StringIO(
        """
id:*up_2009_*
title: *up_2009_*
image_url: *up_2009_*
id_regex: .*up_2009_.*
title_regex: .*up_2009_.*
image_url_regex: .*up_2009_.*
    """
    )

    assert Blurred.parse_blob(buf) == Blurred(
        {
            Blur(Attr.ID_FNMATCH, r"*up_2009_*"),
            Blur(Attr.TITLE_FNMATCH, r"*up_2009_*"),
            Blur(Attr.IMAGE_FNMATCH, r"*up_2009_*"),
            Blur(Attr.ID_REGEX, r".*up_2009_.*"),
            Blur(Attr.TITLE_REGEX, r".*up_2009_.*"),
            Blur(Attr.IMAGE_REGEX, r".*up_2009_.*"),
        }
    )
